/**************************************************************************

    Copyright (C) 2010  Eli Lilly and Company

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include "fingerprint.h"
#include "tbb/task_scheduler_init.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"

static tbb::task_scheduler_init tbb_init;
static tbb::affinity_partitioner tbb_affinity;

static float distance_threshold = 0.2F;

struct Cluster
{
  int cluster_number;
  float dis_from_leader;
  int idx;
};

bool cluster_compare(const Cluster& a, const Cluster& b)
{
   return a.cluster_number < b.cluster_number;
}


static Fingerprint * pool = NULL;
static Cluster *cluster = NULL;
static int cluster_id = 0;
static Fingerprint * centre = NULL;
static int * unselected_items = NULL;
static int unselected_size = 0;
static int n_squeeze = 0;


//for tbb::parallel_for
struct tbb_leader
{
    void operator()(const tbb::blocked_range<int>& r) const
    {
       for( int i=r.begin();i!=r.end();++i)
       {
         int idx = unselected_items[i];
	 if(cluster[idx].cluster_number >= 0) 
	   continue;
	 float dis= tanimoto_distance(*centre,pool[idx]);
	 if(dis > distance_threshold)
	   continue;
	 cluster[idx].cluster_number  = cluster_id;
         cluster[idx].dis_from_leader = dis;
       }
    }
}; 

static int choose_next_centre ()
{
  for(int i=0;i<unselected_size;i++)
  {
    int idx = unselected_items[i];
    if(cluster[idx].cluster_number < 0) //still unselected
    {
      return idx; 
    }
  }
  return -1;
}

static void squeeze_unselected_items ()
{
  n_squeeze++;
  if(n_squeeze < 100)
    return;
  n_squeeze = 0;
  int j=0;
  for(int i=0;i<unselected_size;i++)
  {
    int idx = unselected_items[i];
    if(cluster[idx].cluster_number < 0) //still unselected
    {
      unselected_items[j] = idx;
      j++;
    }
  }
  unselected_size = j;
}

static void usage(int rc)
{
  cerr << "Performs leader clustering on a set of fingerprints\n";
  cerr << "Usage <options> <input_file>\n";
  cerr << "options:\n";
  cerr << " -t <dis>       specify distance threshold (default 0.2)\n";
  cerr << " -A <file>      specify previously selected fingerprints\n";

  exit (rc);
}


int
main (int argc, char ** argv)
{

  char previously_selected_file[200];
  int previous = 0;
  int i;
  for( i=1;(i<argc) && ((argv[i])[0] == '-');i++)
  {
    switch ((argv[i])[1])
    {
      case 'h': usage(0);
      case 't': i++; distance_threshold = atof(argv[i]); break;
      case 'A': i++; previous=1; strcpy(previously_selected_file,argv[i]); break;
      default: 
		cerr<<"\nUnrecognized option "<< argv[i] <<endl<<endl;
		usage(0);
    }
  }
  if(i != (argc-1) ) 
  {
    cerr << "No input fp file !"<<endl;
    usage(0);
  }


  int pool_size=build_fp_pool(argv[i],pool);
  cerr <<"Pool has "<< pool_size << " fingerprints." <<endl;
  cluster=new Cluster[pool_size];
  unselected_items = new int[pool_size];
  unselected_size = pool_size;
  for( i=0;i<pool_size;i++)
  {
    cluster[i].cluster_number = -1;
    cluster[i].dis_from_leader = 0.0F;
    cluster[i].idx = i;
    unselected_items[i] = i;
  }
  
  if(previous)
  {
     Fingerprint * pool2 = NULL;
     int size2 = build_fp_pool(previously_selected_file,pool2);
     cerr <<"Previously selected Pool has "<< size2 << " fingerprints." <<endl;
     for(int j =0; j < size2 && unselected_size > 0;j++)
     {
       centre = &pool2[j];
       tbb::parallel_for( tbb::blocked_range<int>( 0,unselected_size),tbb_leader(),tbb_affinity);    
       squeeze_unselected_items();
     }
     delete [] pool2;
  }
  int icentre =0; 
  while ( (icentre = choose_next_centre()) >= 0 )
  {
    cluster_id ++;
    centre = &pool[icentre];
    tbb::parallel_for( tbb::blocked_range<int>( 0,unselected_size),tbb_leader(),tbb_affinity);     
    squeeze_unselected_items(); 
  }
  cerr << "Find " << cluster_id << " clusters"<<endl;
  std::stable_sort(cluster,cluster+pool_size,cluster_compare);
  for(i=0;i<pool_size;i++)
  {
    cout<<cluster[i].cluster_number <<" "<<pool[cluster[i].idx].id()<<" "<<cluster[i].dis_from_leader<<endl;
  }
  
  delete [] cluster;
  delete [] unselected_items;
  delete [] pool;
  
  return 0;
}
