/**************************************************************************

    Copyright (C) 2010  Eli Lilly and Company

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include "fingerprint.h"
#include "tbb/task_scheduler_init.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"
#include "tbb/parallel_reduce.h"


static tbb::task_scheduler_init tbb_init;
static tbb::affinity_partitioner tbb_affinity;

static Fingerprint * pool = NULL;
static int * nsn = NULL;
static float * min_dis = NULL;
static Fingerprint * centre = NULL;
static int * unselected_items = NULL;
static int unselected_size = 0;
static int n_squeeze = 0;
static int ichoose = 0;

struct tbb_max
{
  float max_dis;
  int choose;
  tbb_max():max_dis(-0.5F),choose(-1) {}
  tbb_max(tbb_max& t,tbb::split) {max_dis = -0.5F;choose=-1;}
  void operator()(const tbb::blocked_range<int>& r)
  {
    for(int i=r.begin();i!=r.end();++i)
    {
      int idx = unselected_items[i];
      if(min_dis[idx] > max_dis)
      {
         max_dis = min_dis[idx];
         choose = idx;
      }
    }
  }
  void join(tbb_max& t) 
  {
    if(t.max_dis > max_dis)
    {
       max_dis = t.max_dis;
       choose = t.choose;
    }
  }
  
};

static int furthest_from_already_selected()
{
  tbb_max tm;
  tbb::parallel_reduce(tbb::blocked_range<int>(0,unselected_size),tm,tbb_affinity);
  ichoose = tm.choose;
  return (ichoose >=0);
}

struct tbb_spread
{
  void operator()(const tbb::blocked_range<int>& r) const
  {
     for( int i=r.begin();i!=r.end();++i)
     {
       int idx = unselected_items[i];
       
       if(min_dis[idx] < -0.5F)
	 continue;
       float dis= tanimoto_distance(*centre,pool[idx]);
       if(dis < min_dis[idx] )
       {
	 min_dis[idx] = dis;
	 nsn[idx] = ichoose;
       }
     }
  }  
};

static void squeeze_unselected_items ()
{
  n_squeeze++;
  if(n_squeeze < 100)
    return;
  n_squeeze = 0;
  int j=0;
  for(int i=0;i<unselected_size;i++)
  {
    int idx = unselected_items[i];
    if(min_dis[idx] > -0.5F) //still unselected
    {
      unselected_items[j] = idx;
      j++;
    }
  }
  unselected_size = j;
}




static void usage(int rc)
{
  cerr << "Performs spread selection on a set of fingerprints\n";
  cerr << "Usage <options> <input_file>\n";
  cerr << "options:\n";
  cerr << " -n <number>    specify number of fingerprints to be selected\n";
  cerr << " -A <file>      specify previously selected fingerprints\n";

  exit (rc);
}

  
int
main (int argc, char ** argv)
{
 
  int  number  = 0;
  char previously_selected_file[200];
  int previous = 0;
  int i;
  for( i=1;(i<argc) && ((argv[i])[0] == '-');i++)
  {
    switch ((argv[i])[1])
    {
      case 'h': usage(0);
      case 'n': i++; number = atoi(argv[i]); break;
      case 'A': i++; previous=1; strcpy(previously_selected_file,argv[i]); break;
      default: 
		cerr<<"\nUnrecognized option "<< argv[i] <<endl<<endl;
		usage(0);
    }
  }
  if(i != (argc-1) ) 
  {
    cerr << "No input fp file !"<<endl;
    usage(0);
  }

  int size=build_fp_pool(argv[i],pool);
  cerr <<"Pool has "<< size << " fingerprints." <<endl;
  
  nsn = new int[size]; 
  min_dis = new float[size];
  unselected_items = new int[size];
  unselected_size = size;
  for(i=0;i<size;i++)
  {
    nsn [i] =0;
    min_dis[i] = 2.0F;
    unselected_items[i] = i;
  }
  int selected =1 ;
  Fingerprint * pool2 = NULL;
  if(previous)
  {
     int size2 = build_fp_pool(previously_selected_file,pool2);
     cerr <<"Previously selected Pool has "<< size2 << " fingerprints." <<endl;
     for(int j =0; j<size2;j++)
     {
       ichoose = -(j+1);
       centre = &pool2[j];
       tbb::parallel_for( tbb::blocked_range<int> (0,unselected_size),tbb_spread(),tbb_affinity);
     }
     furthest_from_already_selected();
     cout <<selected<<" "<< pool[ichoose].id()<<" "<<pool2[-nsn[ichoose]-1].id()<<" "<<min_dis[ichoose]<<endl;
  }
  else
  {
    cout <<selected<<" "<< pool[ichoose].id()<<" "<<"*"<<" "<<"*"<<endl;
  }
  if(number <= 0) number = size;
  while(selected < number)
  {
     min_dis[ichoose] = -1.0F;
     centre = &pool[ichoose];
     tbb::parallel_for( tbb::blocked_range<int> (0,unselected_size),tbb_spread(),tbb_affinity);
     if(!furthest_from_already_selected())
     {
        break;
     }
     ++selected;
     cout <<selected<<" "<< pool[ichoose].id()<<" "<< (nsn[ichoose]<0?pool2[-nsn[ichoose]-1].id():pool[nsn[ichoose]].id())<<" "<<min_dis[ichoose]<<endl;
     squeeze_unselected_items();
  }
  cerr << "Selected " << selected << " fingerprints"<<endl;

  if(pool2)
    delete [] pool2;
  delete [] nsn;
  delete [] min_dis;
  delete [] unselected_items;
  delete [] pool;
  
  return 0;
}




