/**************************************************************************

    Copyright (C) 2010  Eli Lilly and Company

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/


#include "gpu_functions.h"
#include "cuda.h"
#include <thrust/detail/device/cuda/arch.h>
#include <thrust/device_ptr.h>
#include <thrust/sequence.h>
#include <thrust/transform_scan.h>
#include <thrust/scatter.h>
#include <thrust/extrema.h>
#include <thrust/sort.h>
#include <thrust/functional.h>
#include <thrust/device_vector.h>
#include <thrust/reduce.h>
#include <thrust/partition.h>


//Thrust library can be downloaded from http://code.google.com/p/thrust/
//Thrust v1.3.0 is used here 

void * malloc_gpu(unsigned int size)
{
  void *result =0;
  cudaError_t error = cudaMalloc((void**)&result,size);
  if(error)
  {
    cerr << "Cannot allocate GPU memory, size " << size << " reason " << cudaGetErrorString(error) << endl;
    exit(1);
  }
  return result;
}

void * malloc_cpu(unsigned int size)
{
  void *result =0;
  cudaError_t error = cudaMallocHost((void**)&result,size);
  if(error)
  {
    cerr<<"Out of CPU pinned memory"<<endl;
    exit(1);
  }
  return result;
}

void free_gpu(void *p)
{
  cudaFree(p);
}

void free_cpu(void *p)
{
  cudaFreeHost(p);
}

static void checked_cudaMemcpy(void *dst, const void *src, unsigned int  count, enum cudaMemcpyKind kind, int async)
{
  cudaError_t error;
  if(async)
  {
    error = cudaMemcpyAsync(dst,src,count,kind,0);
  }
  else
  {
    error = cudaMemcpy(dst,src,count,kind);
  }
  if(error)
  {
    cerr<<"CUDA Memory copy error: "<<cudaGetErrorString(error)<<endl;
    exit(1);
  }
}

void
cpu_to_gpu(void *gpu, void *cpu,unsigned int size,int async)
{
  checked_cudaMemcpy(gpu,cpu,size,cudaMemcpyHostToDevice,async);
}

void
gpu_to_cpu(void *cpu, void *gpu,unsigned int size,int async)
{
  checked_cudaMemcpy(cpu,gpu,size,cudaMemcpyDeviceToHost,async);
}

void
gpu_to_gpu(void *dst, void *src,unsigned int size,int async)
{
  checked_cudaMemcpy(dst,src,size,cudaMemcpyDeviceToDevice,async);
}

void sync_gpu()
{
  cudaError_t error = cudaThreadSynchronize();
  if(error)
  {
    cerr<<"CUDA Error : "<<cudaGetErrorString(error)<<endl;
    exit(1);
  }
}

void gpu_sequence_int(int *v,int n,int start,int step)
{
  thrust::device_ptr<int> p(v);
  thrust::sequence(p,p+n,start,step);
}


void gpu_fill_float(float *v, int n, float c)
{
  thrust::device_ptr<float> p(v);
  thrust::fill(p,p+n,c);
}

void cpu_sort_by_key(float *dis,int * idx, int n)
{
  thrust::sort_by_key(dis,dis+n,idx);
}


void gpu_sort_by_key(float * dis,int * idx, int n)
{
  thrust::device_ptr<float> d(dis);
  thrust::device_ptr<int> i(idx);
  thrust::sort_by_key(d,d+n,i);
}

/*
__device__ int
popc_gpu(unsigned int a)
{
  a = a - ((a >> 1) & 0x55555555);
  a = (a & 0x33333333) + ((a >> 2) & 0x33333333);
  a = (a + (a >> 4)) & 0x0f0f0f0f;
  a = ((__umul24(a,0x808080) << 1) + a) >> 24;
  return a;
}
*/

__device__ float
tanimoto_distance(const unsigned int *  s_needle,const unsigned int * &g_haystack, int & tid, int &cols)
{
  int n,m,pos=1,common,sum;
  float rc=0.0F,fp=0.0F;
  
  //fixed_size fingerprints
  for(n=s_needle[0];n>0;--n) 
  {
    common=0;
    m=s_needle[pos]+pos+1;
    for(++pos;pos<m;++pos)
    {
      common+=__popc(g_haystack[pos*cols+tid] & s_needle[pos]);
    }
    sum=g_haystack[pos*cols+tid]+s_needle[pos]-common;
    if(sum>0)
    	rc+=(float)common/((float)sum);
    else
       rc+=1.0F;
    ++pos;
    fp+=1.0F;
  }
  return 1.0F - rc/fp;
}


template <bool column_order>
__device__ void
copy_needle_to_share_memory(unsigned int * s_needle,
                            const unsigned int * &g_needle,
			    int & words,int & cols)
{
  //copy the needle data to shared memory
  for(int n=threadIdx.x;n<words;n+=blockDim.x)
  {
    if(column_order)
      s_needle[n]=g_needle[n*cols];
    else
      s_needle[n]=g_needle[n];
  }
  __syncthreads();
}

int gpu_min_int(int * v,int n)
{
  thrust::device_ptr<int> d(v);
  return thrust::reduce(d,d+n,std::numeric_limits<int>::max(),thrust::minimum<int>());
}

void gpu_leader_final_sort(int * index,int * cluster, float * dis,int * index2, int n)
{
  thrust::device_ptr<int> pindex(index);
  thrust::device_ptr<int> pcluster(cluster);
  thrust::device_ptr<float> pdis(dis);
  thrust::device_ptr<int> pindex2(index2);
  gpu_to_gpu(index2,cluster,n*sizeof(int),0);
  thrust::stable_sort_by_key(pcluster,pcluster+n,pindex);
  thrust::stable_sort_by_key(pindex2,pindex2+n,pdis);
}


template <bool column_order> 
__global__ void
tanimoto_distance_row_kernel_leader_threshold(float *dis_from_leader,float threshold,
                 int * cluster_number,int new_id,
                 const unsigned int *g_haystack,
                 int cols,
                 int pool_size,
                 const unsigned int *g_needle,
                 int words)
{
  extern __shared__ unsigned int s_needle[];

  copy_needle_to_share_memory<column_order>(s_needle,g_needle,words,cols);

  int tid=blockDim.x*blockIdx.x+threadIdx.x;
  float d; 
  while(tid < pool_size)
  {
    if(cluster_number[tid] < 0)
    {
      d=tanimoto_distance(s_needle,g_haystack,tid,cols); 
      if( d <= threshold)
      {
      	cluster_number[tid]=new_id;
    	dis_from_leader[tid]=d;
      }
    }
    tid += blockDim.x * gridDim.x;
  }
}

void
gpu_tanimoto_distance_row_leader_threshold(float *dis_from_leader, float threshold,
                        int* cluster_number,int new_id,
                        unsigned int *g_haystack,
                        int cols,int pool_size,
                        unsigned int *g_needle,int words,int column_order)
{
  const size_t BLOCK_SIZE = 256;
  if(column_order)
  {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(tanimoto_distance_row_kernel_leader_threshold<true>,BLOCK_SIZE,words*sizeof(unsigned int));
  const size_t NUM_BLOCKS = std::min(MAX_BLOCKS, (pool_size+(BLOCK_SIZE-1) )/BLOCK_SIZE);
  tanimoto_distance_row_kernel_leader_threshold<true><<<NUM_BLOCKS,BLOCK_SIZE,words*sizeof(unsigned int)>>>(
      dis_from_leader,threshold,cluster_number,new_id,g_haystack,cols,pool_size,g_needle,words);
  }
  else
  {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(tanimoto_distance_row_kernel_leader_threshold<false>,BLOCK_SIZE,words*sizeof(unsigned int));
  const size_t NUM_BLOCKS = std::min(MAX_BLOCKS, (pool_size+(BLOCK_SIZE-1) )/BLOCK_SIZE);
  tanimoto_distance_row_kernel_leader_threshold<false><<<NUM_BLOCKS,BLOCK_SIZE,words*sizeof(unsigned int)>>>(
      dis_from_leader,threshold,cluster_number,new_id,g_haystack,cols,pool_size,g_needle,words);
  }
}

struct Less_Than_Zero
{
  public:
    __host__ __device__
    bool operator()(const int &x)
    {
      return x<0;
    }
};

struct Not_Less_Than_Zero
{
  public:
    __host__ __device__
    bool operator()(const int &x)
    {
      return x>=0;
    }
};

int 
gpu_squeeze_pool_leader(unsigned int * data,int * index, 
     int * cluster, float * dis, unsigned int * scan,
     int * index2,float * dis2, 
     int n,int words,int ceil,int shifted)
{
  thrust::device_ptr<unsigned int> pdata(data);
  thrust::device_ptr<int> pindex(index);
  thrust::device_ptr<int> pindex2(index2);
  thrust::device_ptr<int> pcluster(cluster);
  thrust::device_ptr<float> pdis(dis);
  thrust::device_ptr<float> pdis2(dis2);
  thrust::device_ptr<unsigned int> s(scan);
  thrust::transform_inclusive_scan(pcluster,pcluster+n,s+1,Less_Than_Zero(),thrust::plus<unsigned int>());
  int res=s[n];
  if(shifted)
  {
    for(int i=0;i<words;i++)
      thrust::scatter_if(pdata+ceil*i,pdata+ceil*i+n,s,pcluster,pdata+ceil*(i-1),Less_Than_Zero());
  }
  else
  {
    for(int i=words-1;i>=0;i--)
      thrust::scatter_if(pdata+ceil*i,pdata+ceil*i+n,s,pcluster,pdata+ceil*(i+1),Less_Than_Zero());
  }
  thrust::scatter_if(pindex,pindex+n,s,pcluster,pindex2,Less_Than_Zero());

  thrust::transform_inclusive_scan(pcluster,pcluster+n,s+1,Not_Less_Than_Zero(),thrust::plus<unsigned int>());
  thrust::scatter_if(pindex,pindex+n,s,pcluster,pindex2+res,Not_Less_Than_Zero());
  gpu_to_gpu(index,index2,n*sizeof(int),0);
  thrust::scatter_if(pdis,pdis+n,s,pcluster,pdis2,Not_Less_Than_Zero());
  gpu_to_gpu(dis+res,dis2,(n-res)*sizeof(float),0);
  thrust::scatter_if(pcluster,pcluster+n,s,pcluster,pindex2,Not_Less_Than_Zero());
  gpu_to_gpu(cluster+res,index2,(n-res)*sizeof(float),0);
  return res;
}

template <bool column_order> 
__global__ void
tanimoto_distance_row_kernel_spread(float *min_dis, int * nsn, int nsn_id,
                 const unsigned int *g_haystack,
                 int cols,
                 int pool_size,
                 const unsigned int *g_needle,
                 int words)
{
  extern __shared__ unsigned int s_needle[];

  copy_needle_to_share_memory<column_order>(s_needle,g_needle,words,cols);

  int tid=blockDim.x*blockIdx.x+threadIdx.x;
  float d,mind; 
  while(tid < pool_size)
  {
    mind=min_dis[tid];
    if(mind > -0.5F ) //not be selected yet
    {
      d=tanimoto_distance(s_needle,g_haystack,tid,cols);
      if(d < mind)
      {
        nsn[tid]=nsn_id;
        min_dis[tid]=d;
      }
    }
    tid += blockDim.x * gridDim.x;
  }
}

void
gpu_tanimoto_distance_row_spread(float *min_dis,int * nsn, int nsn_id,
                        unsigned int *g_haystack,
                        int cols,int pool_size,
                        unsigned int *g_needle,int words,int column_order)
{
  const size_t BLOCK_SIZE = 256;
  if(column_order)
  {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(tanimoto_distance_row_kernel_spread<true>,BLOCK_SIZE,words*sizeof(unsigned int));
  const size_t NUM_BLOCKS = std::min(MAX_BLOCKS, (pool_size+(BLOCK_SIZE-1) )/BLOCK_SIZE);
  tanimoto_distance_row_kernel_spread<true><<<NUM_BLOCKS,BLOCK_SIZE,words*sizeof(unsigned int)>>>(min_dis,nsn,nsn_id,g_haystack,cols,pool_size,g_needle,words);
  }
  else
  {
    const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(tanimoto_distance_row_kernel_spread<false>,BLOCK_SIZE,words*sizeof(unsigned int));
  const size_t NUM_BLOCKS = std::min(MAX_BLOCKS, (pool_size+(BLOCK_SIZE-1) )/BLOCK_SIZE);
  tanimoto_distance_row_kernel_spread<false><<<NUM_BLOCKS,BLOCK_SIZE,words*sizeof(unsigned int)>>>(min_dis,nsn,nsn_id,g_haystack,cols,pool_size,g_needle,words);
  }
}

int gpu_max_index_float(float *dis,int n)
{
  thrust::device_ptr<float> d(dis);
  return thrust::max_element(d,d+n)-d;
}


struct Not_Selected 
{
  public:
    __host__ __device__
    bool operator()(const float &x)
    {
      return x > -0.5F;
    }
};

int 
gpu_squeeze_pool_spread(unsigned int * data,int * index, 
     int * nsn, float * dis, unsigned int * scan,
     int * index2,float * dis2, 
     int n,int words,int ceil,int shifted)
{
  thrust::device_ptr<unsigned int> pdata(data);
  thrust::device_ptr<int> pindex(index);
  thrust::device_ptr<int> pindex2(index2);
  thrust::device_ptr<int> pnsn(nsn);
  thrust::device_ptr<float> pdis(dis);
  thrust::device_ptr<float> pdis2(dis2);
  thrust::device_ptr<unsigned int> s(scan);
  thrust::transform_inclusive_scan(pdis,pdis+n,s+1,Not_Selected(),thrust::plus<unsigned int>());
  int res=s[n];
  if(shifted)
  {
    for(int i=0;i<words;i++)
      thrust::scatter_if(pdata+ceil*i,pdata+ceil*i+n,s,pdis,pdata+ceil*(i-1),Not_Selected());
  }
  else
  {
    for(int i=words-1;i>=0;i--)
      thrust::scatter_if(pdata+ceil*i,pdata+ceil*i+n,s,pdis,pdata+ceil*(i+1),Not_Selected());
  }
  thrust::scatter_if(pindex,pindex+n,s,pdis,pindex2,Not_Selected());
  gpu_to_gpu(index,index2,res*sizeof(int),0);

  thrust::scatter_if(pnsn,pnsn+n,s,pdis,pindex2,Not_Selected());
  gpu_to_gpu(nsn,index2,res*sizeof(int),0);

  thrust::scatter_if(pdis,pdis+n,s,pdis,pdis2,Not_Selected());
  gpu_to_gpu(dis,dis2,res*sizeof(float),0);
  return res;
}



