/**************************************************************************

    Copyright (C) 2010  Eli Lilly and Company

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include "gpu_haystack_leader.h"

GPU_Haystack_Leader::GPU_Haystack_Leader()
{
	
}
		
GPU_Haystack_Leader::~GPU_Haystack_Leader()
{
  clear();	
}

void
GPU_Haystack_Leader::clear()
{
  if(size()>0)
  {
    free_gpu(_d_cluster_number);
    free_gpu(_d_dis_from_leader);
    free_cpu(_cluster_number);
    free_cpu(_dis_from_leader);
    free_cpu(_index);
  }
}

int GPU_Haystack_Leader::init(Fingerprint * pool, int pool_size)
{
  clear();

  if(!GPU_Haystack::init(pool,pool_size))
    return 0;
  
  _d_cluster_number = (int *)malloc_gpu(_mols * sizeof(int));
  _d_dis_from_leader = (float *)malloc_gpu(_mols * sizeof(float));
  _cluster_number = (int *) malloc_cpu(_mols * sizeof(int));
  _dis_from_leader = (float *)malloc_cpu( _mols * sizeof(float));
  _index = (int *) malloc_cpu(_mols * sizeof(int));
  
  //initilize the _d_cluster_number to a sequence (-mols , -1)
  gpu_sequence_int(_d_cluster_number,_mols,-_mols,1);
  _cluster_id=0;
  return 1; 
}

int 
GPU_Haystack_Leader::squeeze()
{
  sync_gpu();
  int res=gpu_squeeze_pool_leader(_d_data+_start+_shifted*_ceil,_d_index+_start,
           _d_cluster_number+_start,_d_dis_from_leader+_start,_d_scan,_d_index_tmp,_d_dis_tmp,range(),_words,_ceil,_shifted);
  _shifted=1-_shifted;
  _end=_start+res;
  gpu_sequence_int(_d_cluster_number+_start,range(),_start-_mols,1);
  return res;
}

int
GPU_Haystack_Leader::choose_next_centre(int & icentre)
{
  if(range() <= 0) 
    return 0;
  icentre = gpu_min_int(_d_cluster_number+_start,range())+_mols;
  return icentre < _mols;
}

void GPU_Haystack_Leader::do_leader(int icentre,float threshold)
{
  _cluster_id++;
  set_start(icentre);
  //start from a memory aligned address
  int ceil_start=_start - _start % 32;
  gpu_tanimoto_distance_row_leader_threshold(_d_dis_from_leader+ceil_start, threshold, _d_cluster_number + ceil_start, _cluster_id,
     _d_data+ceil_start+_shifted*_ceil,_ceil,_end-ceil_start,_d_data+icentre+_shifted*_ceil,_words,1);
  set_start(icentre+1);
}

int GPU_Haystack_Leader::final_process()
{
  gpu_leader_final_sort(_d_index,_d_cluster_number,_d_dis_from_leader,_d_index_tmp, _mols);
  gpu_to_cpu(_cluster_number,_d_cluster_number,_mols*sizeof(int),0);
  gpu_to_cpu(_dis_from_leader,_d_dis_from_leader,_mols*sizeof(float),0);
  gpu_to_cpu(_index,_d_index,_mols*sizeof(int),0);
  return _cluster_id;  
}


void 
GPU_Haystack_Leader::do_previously_selected (
               GPU_Needle & needle,float threshold)
{
  int ceil_start=_start - _start % 32;
  gpu_tanimoto_distance_row_leader_threshold(_d_dis_from_leader+ceil_start, threshold, _d_cluster_number + ceil_start, 0,
	_d_data+ceil_start+_shifted*_ceil,_ceil,_end-ceil_start,needle.get_ptr(),needle.words(),0);
}

  
int GPU_Haystack_Leader::leader(Fingerprint * pool, float threshold)
{
  int icentre=0;
  int n_squeeze=0;
  while (choose_next_centre (icentre))
  {
    do_leader(icentre,threshold); 
    n_squeeze++;
    if(n_squeeze == 1000 )
    {
      n_squeeze=0;
      squeeze();
    }
  }
  int clusters=final_process(); 
  for(int i=0;i<_mols;i++)
  {
    cout<<_cluster_number[i] <<" "<<pool[_index[i]].id()<<" "<<_dis_from_leader[i]<<endl;
  }
  return clusters;
}


static void usage(int rc)
{
  cerr << "Performs leader clustering on a set of fingerprints\n";
  cerr << "Usage <options> <input_file>\n";
  cerr << "options:\n";
  cerr << " -t <dis>       specify distance threshold (default 0.2)\n";
  cerr << " -A <file>      specify previously selected fingerprints\n";

  exit (rc);
}


int
main (int argc, char ** argv)
{

  float distance_threshold = 0.2F;
  char previously_selected_file[200];
  int previous = 0;
  int i;
  for( i=1;(i<argc) && ((argv[i])[0] == '-');i++)
  {
    switch ((argv[i])[1])
    {
      case 'h': usage(0);
      case 't': i++; distance_threshold = atof(argv[i]); break;
      case 'A': i++; previous=1; strcpy(previously_selected_file,argv[i]); break;
      default: 
		cerr<<"\nUnrecognized option "<< argv[i] <<endl<<endl;
		usage(0);
    }
  }
  if(i != (argc-1) ) 
  {
    cerr << "No input fp file !"<<endl;
    usage(0);
  }

  Fingerprint * pool = NULL;

  int size=build_fp_pool(argv[i],pool);
  cerr <<"Pool has "<< size << " fingerprints." <<endl;
  
  GPU_Haystack_Leader haystack;
  haystack.init(pool,size);

  if(previous)
  {
     Fingerprint * pool2 = NULL;
     int size2 = build_fp_pool(previously_selected_file,pool2);
     cerr <<"Previously selected Pool has "<< size2 << " fingerprints." <<endl;
     GPU_Needle needle;
     for(int j =0; j<size2;j++)
     {
       needle.init(pool2[j]);
       haystack.do_previously_selected(needle,distance_threshold);
     }
     haystack.squeeze();
     delete [] pool2;
  }
 
  int n=haystack.leader(pool,distance_threshold);
  cerr << "Find " << n << " clusters"<<endl;
  delete [] pool;
  
  return 0;
}
