/**************************************************************************

    Copyright (C) 2010  Eli Lilly and Company

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include "gpu_haystack_spread.h"

GPU_Haystack_Spread::GPU_Haystack_Spread()
{
	
}
		
GPU_Haystack_Spread::~GPU_Haystack_Spread()
{
  clear();	
}

void
GPU_Haystack_Spread::clear()
{
  if(size()>0)
  {
    free_gpu(_d_min_dis);
    free_gpu(_d_nsn);
  }
}

int GPU_Haystack_Spread::init(Fingerprint * pool, int pool_size)
{
  clear();

  if(!GPU_Haystack::init(pool,pool_size))
    return 0;

  _d_min_dis = (float *)malloc_gpu(_mols * sizeof(float));
  _d_nsn = (int *)malloc_gpu(_mols * sizeof(int));

  gpu_fill_float(_d_min_dis,_mols,2.0F); //initilize all min_dis to 2.0F (large than any possible distance ( [0,1] )
  
  return 1;
}

void GPU_Haystack_Spread::do_spread(int offset,int idx)
{
  //start from a memory aligned address
  int ceil_start=_start - _start % 32;
  
  //exclude the current selected molecule by assign it's min_dis to -1.0F
  float minusone = -1.0F;
  cpu_to_gpu(_d_min_dis + _start + offset,&minusone, sizeof(float),0);

  gpu_tanimoto_distance_row_spread(_d_min_dis+ceil_start, _d_nsn + ceil_start, idx,
     _d_data+ceil_start+_shifted*_ceil,_ceil,_end-ceil_start,_d_data+_start+offset+_shifted*_ceil,_words,1);
}

void 
GPU_Haystack_Spread::do_previously_selected (
               GPU_Needle & needle,int idx)
{
  int ceil_start=_start - _start % 32;
  gpu_tanimoto_distance_row_spread(_d_min_dis+ceil_start, _d_nsn + ceil_start, idx,
	_d_data+ceil_start+_shifted*_ceil,_ceil,_end-ceil_start,needle.get_ptr(),needle.words(),0);
}

int GPU_Haystack_Spread::furthest_from_already_selected(int &offset,int &idx, int & nsn, float & dis )
{
  offset=gpu_max_index_float(_d_min_dis+_start,range());
  
  gpu_to_cpu(&dis,_d_min_dis+_start+offset,sizeof(float),0);
  if(dis < -0.5F )// this is already selected!  
    return 0;
  gpu_to_cpu(&idx,_d_index+_start+offset,sizeof(int),0);
  gpu_to_cpu(&nsn,_d_nsn+_start+offset,sizeof(int),0);
  return 1 ;
}

int GPU_Haystack_Spread::squeeze()
{
  sync_gpu();
  int res=gpu_squeeze_pool_spread(_d_data+_start+_shifted*_ceil,_d_index+_start,
           _d_nsn+_start,_d_min_dis+_start,_d_scan,_d_index_tmp,_d_dis_tmp,range(),_words,_ceil,_shifted);
  _shifted=1-_shifted;
  _end=_start+res;
  return res;
}


static void usage(int rc)
{
  cerr << "Performs spread selection on a set of fingerprints\n";
  cerr << "Usage <options> <input_file>\n";
  cerr << "options:\n";
  cerr << " -n <number>    specify number of fingerprints to be selected\n";
  cerr << " -A <file>      specify previously selected fingerprints\n";

  exit (rc);
}

  
int
main (int argc, char ** argv)
{
 
  int  number  = 0;
  char previously_selected_file[200];
  int previous = 0;
  int i;
  for( i=1;(i<argc) && ((argv[i])[0] == '-');i++)
  {
    switch ((argv[i])[1])
    {
      case 'h': usage(0);
      case 'n': i++; number = atoi(argv[i]); break;
      case 'A': i++; previous=1; strcpy(previously_selected_file,argv[i]); break;
      default: 
		cerr<<"\nUnrecognized option "<< argv[i] <<endl<<endl;
		usage(0);
    }
  }
  if(i != (argc-1) ) 
  {
    cerr << "No input fp file !"<<endl;
    usage(0);
  }

  Fingerprint * pool = NULL;

  int size=build_fp_pool(argv[i],pool);
  cerr <<"Pool has "<< size << " fingerprints." <<endl;
  
  GPU_Haystack_Spread haystack;
  haystack.init(pool,size);

  int nsn; 
  float dis;
  int offset = 0;
  int ichoose = 0;
  int selected =1 ;
  Fingerprint * pool2 = NULL;
  if(previous)
  {
     int size2 = build_fp_pool(previously_selected_file,pool2);
     cerr <<"Previously selected Pool has "<< size2 << " fingerprints." <<endl;
     GPU_Needle needle;
     for(int j =0; j<size2;j++)
     {
       needle.init(pool2[j]);
       haystack.do_previously_selected(needle,-(j+1));
     }
     haystack.furthest_from_already_selected(offset,ichoose,nsn,dis);
     cout <<selected<<" "<< pool[ichoose].id()<<" "<<pool2[-nsn-1].id()<<" "<<dis<<endl;
  }
  else
  {
    cout <<selected<<" "<< pool[ichoose].id()<<" "<<"*"<<" "<<"*"<<endl;
  }
  int n_squeeze=1;
  if(number <= 0) number = size;
  while(selected < number)
  {
     haystack.do_spread(offset,ichoose);
     ++n_squeeze;
    if(n_squeeze == 1000 )
    {
      n_squeeze=0;
      haystack.squeeze();
    }
     if(!haystack.furthest_from_already_selected(offset,ichoose,nsn,dis))
     {
        break;
     }
     ++selected;
     cout <<selected<<" "<< pool[ichoose].id()<<" "<< (nsn<0?pool2[-nsn-1].id():pool[nsn].id())<<" "<<dis<<endl;
  }
  cerr << "Selected " << selected << " fingerprints"<<endl;

  if(pool2)
    delete [] pool2;
  delete [] pool;
  
  return 0;
}




